package smtml

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Expr.PrinterSource

Sourceval pp_expr : t Fmt.t

pp_expr is a printer for expressions that can be parsed back.

Sourceval pp_query : t list Fmt.t

pp_query is a printer for a list of expressions in query mode.